import asyncio
from asyncio.base_events import BaseEventLoop
from grpc import Server
from grpc.experimental import aio
from grpc_reflection.v1alpha import reflection
from grpc_health.v1 import health, health_pb2_grpc
from grpc_health.v1.health_pb2 import _HEALTH, HealthCheckResponse
from delphai_utils.logging import logging
from google.protobuf.descriptor import FileDescriptor
from delphai_utils.gateway import start_gateway
import signal

shutdown_event = asyncio.Event()


def create_grpc_server(descriptor: FileDescriptor):
  server = aio.server()
  server.__dict__['descriptor'] = descriptor
  server.add_insecure_port('[::]:8080')
  health_servicer = health.HealthServicer(experimental_non_blocking=True)
  health_pb2_grpc.add_HealthServicer_to_server(health_servicer, server)
  services = descriptor.services_by_name.keys()
  service_full_names = list(map(lambda service: descriptor.services_by_name[service].full_name, services))
  service_names = (
    *service_full_names,
    _HEALTH.full_name,
    reflection.SERVICE_NAME,
  )
  reflection.enable_server_reflection(service_names, server)
  return server


def start_server(server: Server, gateway: bool = True):
  logging.info('starting grpc server...')
  loop = asyncio.get_event_loop()
  loop.run_until_complete(server.start())
  logging.info('started grpc server on port 8080')
  try:
    if gateway:
      gateway = start_gateway(server.__dict__['descriptor'])
      loop.create_task(gateway)
    loop.create_task(server.wait_for_termination())
    loop.run_forever()
  except KeyboardInterrupt:
    logging.info('stopped server (keyboard interrupt)')


from delphai_utils.validator import validate as _validate, ValidationFailed
from grpc import StatusCode


async def validate(descriptor, value, context):
  try:
    validated = _validate(descriptor)(value)
    return
  except ValidationFailed as ex:
    args = list(ex.args)
    errors = list(map(lambda err: err.replace('p.', ''), args))
    await context.abort(StatusCode.INVALID_ARGUMENT, ', '.join(errors))


from base64 import b64decode
import json


def get_user(context):
  metadata = dict(context.invocation_metadata())
  user64 = metadata['x-delphai-user']
  user = b64decode(user64)
  return json.loads(user)



from multiprocessing import current_process
from threading import Thread
from time import perf_counter, sleep
from abc import ABC, abstractmethod


class Cacher(ABC):
  """
    Implements an autorenewable cache for some data
    """
  name = 'Abstract cache'
  renew_interval = None
  timeout = 60
  _data = None

  def __init__(self):
    Thread(target=cache_renewer_worker, daemon=True, args=[self]).start()

  def get_data(self):
    """
        Returnes data from cache.
        """
    start = perf_counter()
    while self._data is None:
      if perf_counter() - start > self.timeout:
        raise Exception(f'Data acqisition timeout for cache "{self.name}"')
      sleep(0.2)
    return self._data

  def is_ready(self):
    """Returns True if data acquired and ready to use"""
    return self._data is not None

  @abstractmethod
  def get_data_func(self):
    """Must return something that will be stored into _data"""
    pass


def cache_renewer_worker(cacher: Cacher):
  cacher._data = cacher.get_data_func()
  if cacher.renew_interval:
    while True:
      sleep(cacher.renew_interval)
      cacher._data = cacher.get_data_func()


CACHES_BY_PID = {}  # process id -> cache name -> Cacher

REGISTERED_CACHERS = {}  # cache name -> (Cacher class, autoload)


def register_cacher(cacher_class, autoload: bool = False):
  name = cacher_class.name
  REGISTERED_CACHERS[name] = (cacher_class, autoload)
  if autoload:
    get_cache(name)


def get_cache(name) -> Cacher:
  pid = current_process().pid
  if pid not in CACHES_BY_PID:
    names2init = set(k for k, v in REGISTERED_CACHERS.items() if v[1])
    names2init.add(name)
    CACHES_BY_PID[pid] = {n: REGISTERED_CACHERS[n][0]() for n in names2init}
  elif name not in CACHES_BY_PID[pid]:
    CACHES_BY_PID[pid][name] = REGISTERED_CACHERS[name][0]()

  return CACHES_BY_PID[pid][name]



import os, signal, threading, sys
from delphai_utils.logging import logging


# Disable
def block_print():
  sys.stdout = open(os.devnull, 'w')


# Restore
def enable_print():
  sys.stdout = sys.__stdout__


def _signal_handler(sig, frame):
  logging.info('stopping server...')
  sys.exit(0)


def wait_for_exit():
  signal.signal(signal.SIGINT, _signal_handler)
  forever = threading.Event()
  forever.wait()


def generate_default_message(message_descriptor, lvls: int = 10):
  """
  Returns descriptive template for message.
  """
  if lvls == 0:
    return '...'
  res = {}
  for fld in message_descriptor.fields:
    if fld.message_type:
      fld_content = generate_default_message(fld.message_type, lvls - 1)
    else:
      fld_content = fld.default_value
    if fld.default_value == []:
      if fld.message_type:
        fld_content = [
          fld_content,
        ]
      elif fld.type == 9:  # string
        fld_content = [
          '',
        ]
      elif fld.type == 2:  # float
        fld_content = [
          0.,
        ]
      elif fld.type == 8:  # boolean
        fld_content = [
          False,
        ]
      elif fld.type in [3, 5]:  # int32, int64
        fld_content = [
          0,
        ]
    res[fld.name] = fld_content
  return res


import asyncio
from asyncio.base_events import BaseEventLoop
import threading
from typing import List
from google.protobuf.descriptor import Descriptor, FieldDescriptor, FileDescriptor, MethodDescriptor, ServiceDescriptor
from google.protobuf.json_format import MessageToDict
from google.protobuf.message import Message
from grpc import Channel, Server, StatusCode
from delphai_utils.logging import logging
from google.protobuf.descriptor_pb2 import MethodOptions
from google.api.http_pb2 import HttpRule
from starlette.applications import Starlette
from starlette.responses import JSONResponse
from starlette.routing import Route
from starlette.exceptions import HTTPException
from starlette.middleware.base import BaseHTTPMiddleware
from grpc.experimental.aio import insecure_channel, AioRpcError
from starlette.middleware.cors import CORSMiddleware
import os
from google.protobuf import symbol_database
from hypercorn.config import Config
from hypercorn.asyncio import serve
from grpc import StatusCode
from urllib.parse import urlparse
from time import perf_counter
from starlette_prometheus import metrics, PrometheusMiddleware
from delphai_utils.utils import generate_default_message
supported_methods = ['get', 'put', 'post', 'delete', 'patch']

status_map = {
  StatusCode.OK: 200,
  StatusCode.CANCELLED: 499,
  StatusCode.UNKNOWN: 500,
  StatusCode.INVALID_ARGUMENT: 400,
  StatusCode.DEADLINE_EXCEEDED: 504,
  StatusCode.NOT_FOUND: 404,
  StatusCode.ALREADY_EXISTS: 409,
  StatusCode.PERMISSION_DENIED: 403,
  StatusCode.UNAUTHENTICATED: 401,
  StatusCode.RESOURCE_EXHAUSTED: 429,
  StatusCode.FAILED_PRECONDITION: 412,
  StatusCode.ABORTED: 499,
  StatusCode.OUT_OF_RANGE: 416,
  StatusCode.UNIMPLEMENTED: 404,
  StatusCode.INTERNAL: 418,
  StatusCode.UNAVAILABLE: 503,
  StatusCode.DATA_LOSS: 420
}


class AccessLogMiddleware(BaseHTTPMiddleware):
  async def dispatch(self, request, call_next):
    start = perf_counter()
    response = await call_next(request)
    if response.status_code < 400:
      path = urlparse(str(request.url)).path
      end = perf_counter()
      elapsed = round((end - start) * 1000, 2)
      logging.info(f'[{response.status_code}] {request.method} {path} [{elapsed}ms]')
    return response


async def process_grpc_request(descriptor: FileDescriptor, service_name: str, method: str, input: Message,
                               channel: Channel):
  service = descriptor.services_by_name[service_name]
  method_descriptor = service.methods_by_name[method]
  input_prototype = symbol_database.Default().GetPrototype(method_descriptor.input_type)
  output_prototype = symbol_database.Default().GetPrototype(method_descriptor.output_type)
  method_callable = channel.unary_unary(f'/{service.full_name}/{method}',
                                        request_serializer=input_prototype.SerializeToString,
                                        response_deserializer=output_prototype.FromString)
  response = await method_callable(input_prototype(**input))
  return response


async def http_exception(request, exc):
  if 'favicon.ico' in str(request.url):
    detail = 'not found'
    status_code = 404
  else:
    path = urlparse(str(request.url)).path
    logging.error(f'[{exc.status_code}] {request.method} {path} - {exc.detail}')
    detail = exc.detail
    status_code = exc.status_code
  return JSONResponse({'detail': detail, 'status': exc.status_code}, status_code=status_code)


def get_http_handlers(descriptor: FileDescriptor, service_name: str, method: str, channel: Channel):
  async def method_get_handler(request):
    service = descriptor.services_by_name[service_name]
    method_descriptor = service.methods_by_name[method]
    input = generate_default_message(method_descriptor.input_type)
    output = generate_default_message(method_descriptor.output_type)

    function_name = f'{service.full_name}.{method}'
    return JSONResponse({'function_name': function_name, 'input': input, 'output': output})

  async def request_handler(request):
    body = {}
    if len(await request.body()) > 0:
      body = await request.json()
    input = {**request.path_params, **request.query_params, **body}
    try:
      raw_output = await process_grpc_request(descriptor, service_name, method, input, channel)
      output = MessageToDict(raw_output, preserving_proto_field_name=True)
      return JSONResponse(output)
    except AioRpcError as ex:
      detail = ex.details()
      grpc_status = ex.code()
      http_status_code = status_map[grpc_status]
      raise HTTPException(http_status_code, detail=detail)
    except Exception as ex:
      detail = str(ex).replace('\n', ' ')
      http_status_code = 500
      raise HTTPException(http_status_code, detail=detail)

  return method_get_handler, request_handler


async def start_gateway(descriptor: FileDescriptor, port: int = 7070):
  try:
    logging.info('starting gateway...')
    debug = 'DELPHAI_ENVIRONMENT' in os.environ and os.environ['DELPHAI_ENVIRONMENT'] == 'development'
    app = Starlette(debug=debug)
    app.add_exception_handler(HTTPException, http_exception)
    app.add_middleware(AccessLogMiddleware)
    app.add_middleware(PrometheusMiddleware)
    app.add_middleware(CORSMiddleware,
                       allow_origins=['*'],
                       allow_methods=['*'],
                       allow_headers=['*'],
                       allow_credentials=True)
    app.add_exception_handler(HTTPException, http_exception)
    app.add_route("/metrics/", metrics)
    channel = insecure_channel('localhost:8080')
    config = Config()
    config.bind = [f'0.0.0.0:{port}']
    for service_name in descriptor.services_by_name.keys():
      service_handler: ServiceDescriptor = descriptor.services_by_name[service_name]
      if service_handler.full_name.startswith('grpc.'):
        logging.info(f'skipping service {service_handler.name}')
      else:
        logging.info(f'processing service {service_handler.name}')
        for key in service_handler.methods_by_name.keys():
          method_get_handler, request_handler = get_http_handlers(descriptor, service_handler.name, key, channel)
          service_name = key[1:].split('/')[0].split('.')[-1]
          app.add_route(f'/{service_handler.full_name}.{key}', route=method_get_handler, methods=['get'])
          app.add_route(f'/{service_handler.full_name}.{key}', route=request_handler, methods=['post'])
          logging.info(f'  processing {key}')
          method_descriptor: MethodDescriptor = service_handler.methods_by_name[key]
          method_options: MethodOptions = method_descriptor.GetOptions()
          fields: List(FieldDescriptor) = method_options.ListFields()
          for field in fields:
            if field[0].full_name == 'google.api.http':
              http_rule: HttpRule = field[1]
              for supported_method in supported_methods:
                http_path = getattr(http_rule, supported_method)
                if http_path != '':
                  app.add_route(http_path, route=request_handler, methods=[supported_method])
    logging.info(f'started gateway on port {port}')
    return await serve(app, config)
  except Exception as ex:
    logging.error(str(ex))
